@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2009, Regents of the University of Alaska
+Copyright (c) 2013, Regents of the University of Alaska
 
 All rights reserved.
 
@@ -46,7 +46,7 @@ This code was developed by Dan Stahlke for the Geographic Information Network of
 
 #include "config.h"
 
-#include <math.h>
+#include <cmath>
 #include <stdio.h>
 
 #if HAVE_STDINT_H
