@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2009, Regents of the University of Alaska
+Copyright (c) 2013, Regents of the University of Alaska
 
 All rights reserved.
 
@@ -27,6 +27,9 @@ This code was developed by Dan Stahlke for the Geographic Information Network of
 
 
 #include <boost/lexical_cast.hpp>
+#include <boost/foreach.hpp>
+#include <map>
+#include <vector>
 
 #include "common.h"
 #include "polygon.h"
@@ -51,13 +54,6 @@ This code was developed by Dan Stahlke for the Geographic Information Network of
 #define WKB_BYTE_ORDER wkbXDR
 #endif
 
-enum CoordSystem {
-	CS_UNKNOWN,
-	CS_XY,
-	CS_EN,
-	CS_LL
-};
-
 using namespace dangdal;
 
 void usage(const std::string &cmdname) {
@@ -68,72 +64,86 @@ void usage(const std::string &cmdname) {
 	printf("\n");
 	NdvDef::printUsage();
 
-	printf("\
-\n\
-Behavior:\n\
-  -classify                    Output a polygon for each value of an 8-bit band\n\
-                               (default is to generate a single polygon that\n\
-                               surrounds all pixels that don't match\n\
-                               the no-data-value)\n\
-  -b band_id -b band_id ...    Bands to inspect (default is all bands)\n\
-  -invert                      Trace no-data pixels rather than data pixels\n\
-  -erosion                     Erode pixels that don't have two consecutive\n\
-                               neighbors\n\
-  -major-ring                  Take only the biggest outer ring\n\
-  -no-donuts                   Take only top-level rings\n\
-  -min-ring-area val           Drop rings with less than this area\n\
-                               (in square pixels)\n\
-  -dp-toler val                Tolerance for polygon simplification\n\
-                               (in pixels, default is 2.0)\n\
-  -bevel-size                  How much to shave off corners at\n\
-                               self-intersection points\n\
-                               (in pixels, default is 0.1)\n\
-                               (this is done to make geometries that\n\
-                               PostGIS/GEOS/Jump can handle)\n\
-  -pinch-excursions            Remove all the complicated 'mouse bites' that\n\
-                               occur in the outline when lossy compression\n\
-                               has been used (experimental)\n\
-\n\
-Output:\n\
-  -report fn.ppm               Output graphical report of polygons found\n\
-  -mask-out fn.pbm             Output mask of bounding polygon in PBM format\n\
-  -out-cs [xy | en | ll]       Set coordinate system for following outputs\n\
-                               (pixel coords, easting/northing, or lon/lat)\n\
-                               Must be specified before -{wkt,wkb,ogr}-out options\n\
-  -llproj-toler val            Error tolerance for curved lines when\n\
-                               using '-out-cs ll' (in pixels, default is 1.0)\n\
-  -wkt-out fn.wkt              Output polygons in WKT format\n\
-  -wkb-out fn.wkb              Output polygons in WKB format\n\
-  -ogr-out fn.shp              Output polygons using an OGR format\n\
-  -ogr-fmt                     OGR format to use (default is 'ESRI Shapefile')\n\
-                               Must be specified before -ogr-out option\n\
-  -split-polys                 Output several polygons rather than one\n\
-                               multipolygon\n\
-\n\
-Misc:\n\
-  -v                           Verbose\n\
-\n\
-Examples:\n\
-\n\
-Inspect image and output contour of data region:\n\
-gdal_trace_outline raster.tif -nodataval 0 -erosion -out-cs ll -wkt-out outline.wkt\n\
-\n\
-Same as above but polygon actually follows border pixel-by-pixel:\n\
-gdal_trace_outline raster.tif -nodataval 0 -dp-toler 0 -out-cs ll -wkt-out outline.wkt\n\
-\n\
-Output ESRI Shapefile in projection of input image:\n\
-gdal_trace_outline raster.tif -nodataval 0 -erosion -out-cs en -ogr-out outline.shp\n\
-\n\
-Generate one shape for each value in input image:\n\
-gdal_trace_outline raster.tif -classify -out-cs en -ogr-out outline.shp\n\
-\n\
-");
+	printf(
+"\n"
+"Behavior:\n"
+"  -classify                    Output a polygon for each value of an 8-bit band\n"
+"                               (default is to generate a single polygon that\n"
+"                               surrounds all pixels that don't match\n"
+"                               the no-data-value)\n"
+"  -b band_id -b band_id ...    Bands to inspect (default is all bands)\n"
+"  -invert                      Trace no-data pixels rather than data pixels\n"
+"  -erosion                     Erode pixels that don't have two consecutive\n"
+"                               neighbors\n"
+"  -major-ring                  Take only the biggest outer ring\n"
+"  -no-donuts                   Take only top-level rings\n"
+"  -min-ring-area val           Drop rings with less than this area\n"
+"                               (in square pixels)\n"
+"  -containing [xy | en | ll | percent] xval yval\n"
+"                               Take only rings that contain the given point.\n"
+"                               This option can be specified more than once.\n"
+"                               'percent' means interpret coordinate as percent\n"
+"                               of image width/height.\n"
+"  -not-containing [xy | en | ll | percent] xval yval\n"
+"                               Take rings that don't contain the given point.\n"
+"                               This option can be specified more than once.\n"
+"                               'percent' means interpret coordinate as percent\n"
+"                               of image width/height.\n"
+"  -dp-toler val                Tolerance for polygon simplification\n"
+"                               (in pixels, default is 2.0)\n"
+"  -bevel-size                  How much to shave off corners at\n"
+"                               self-intersection points\n"
+"                               (in pixels, default is 0.1)\n"
+"                               (this is done to make geometries that\n"
+"                               PostGIS/GEOS/Jump can handle)\n"
+"  -pinch-excursions            Remove all the complicated 'mouse bites' that\n"
+"                               occur in the outline when lossy compression\n"
+"                               has been used (experimental)\n"
+"\n"
+"Output:\n"
+"  -report fn.ppm               Output graphical report of polygons found\n"
+"  -mask-out fn.pbm             Output mask of bounding polygon in PBM format\n"
+"  -out-cs [xy | en | ll]       Set coordinate system for following outputs\n"
+"                               (pixel coords, easting/northing, or lon/lat)\n"
+"                               Must be specified before -{wkt,wkb,ogr}-out options\n"
+"  -llproj-toler val            Error tolerance for curved lines when\n"
+"                               using '-out-cs ll' (in pixels, default is 1.0)\n"
+"  -wkt-out fn.wkt              Output polygons in WKT format\n"
+"  -wkb-out fn.wkb              Output polygons in WKB format\n"
+"  -ogr-out fn.shp              Output polygons using an OGR format\n"
+"  -ogr-fmt                     OGR format to use (default is 'ESRI Shapefile')\n"
+"                               Must be specified before -ogr-out option\n"
+"  -split-polys                 Output several polygons rather than one\n"
+"                               multipolygon\n"
+"\n"
+"Misc:\n"
+"  -v                           Verbose\n"
+"\n"
+"Examples:\n"
+"\n"
+"Inspect image and output contour of data region:\n"
+"gdal_trace_outline raster.tif -ndv 0 -erosion -out-cs ll -wkt-out outline.wkt\n"
+"\n"
+"Same as above but polygon actually follows border pixel-by-pixel:\n"
+"gdal_trace_outline raster.tif -ndv 0 -dp-toler 0 -out-cs ll -wkt-out outline.wkt\n"
+"\n"
+"Output ESRI Shapefile in projection of input image:\n"
+"gdal_trace_outline raster.tif -ndv 0 -erosion -out-cs en -ogr-out outline.shp\n"
+"\n"
+"Generate one shape for each value in input image:\n"
+"gdal_trace_outline raster.tif -classify -out-cs en -ogr-out outline.shp\n"
+"\n"
+	);
 	exit(1);
 }
 
-Mpoly calc_ring_from_mask(BitGrid &mask, size_t w, size_t h,
-	bool major_ring_only, bool no_donuts,
-	int64_t min_ring_area, double bevel_size);
+enum CoordSystem {
+	CS_UNKNOWN,
+	CS_XY,
+	CS_EN,
+	CS_LL,
+	CS_PERCENT
+};
 
 struct GeomOutput {
 	explicit GeomOutput(CoordSystem _out_cs=CS_UNKNOWN) :
@@ -141,8 +151,11 @@ struct GeomOutput {
 		wkt_fh(NULL),
 		wkb_fh(NULL),
 		ogr_ds(NULL),
-		ogr_layer(NULL)
-	{ }
+		ogr_layer(NULL),
+		class_fld_idx(-1)
+	{
+		for(int i=0; i<4; i++) color_fld_idx[i] = -1;
+	}
 
 	CoordSystem out_cs;
 
@@ -160,6 +173,24 @@ struct GeomOutput {
 	int color_fld_idx[4];
 };
 
+struct ContainingOption {
+	ContainingOption() : x(0), y(0), cs(CS_UNKNOWN), wanted_point(false) { }
+	double x, y;
+	CoordSystem cs;
+	bool wanted_point;
+};
+
+Mpoly take_largest_ring(const Mpoly &mp_in);
+
+Mpoly remove_holes(const Mpoly &mp_in);
+
+Mpoly containment_filters(
+	const Mpoly &mp_in,
+	const std::vector<ContainingOption> &containing_options,
+	const GeoRef &georef,
+	DebugPlot *dbuf
+);
+
 int main(int argc, char **argv) {
 	const std::string cmdname = argv[0];
 	if(argc == 1) usage(cmdname);
@@ -175,7 +206,8 @@ int main(int argc, char **argv) {
 	std::vector<GeomOutput> geom_outputs;
 	std::string mask_out_fn;
 	bool major_ring_only = 0;
-	bool no_donuts = 0;
+	bool trace_no_donuts = 0;
+	bool output_no_donuts = 0;
 	int64_t min_ring_area = 0;
 	double reduction_tolerance = 2;
 	bool do_erosion = 0;
@@ -183,6 +215,7 @@ int main(int argc, char **argv) {
 	double llproj_toler = 1;
 	double bevel_size = .1;
 	bool do_pinch_excursions = 0;
+	std::vector<ContainingOption> containing_options;
 
 	GeoOpts geo_opts = GeoOpts(arg_list);
 	NdvDef ndv_def = NdvDef(arg_list);
@@ -242,7 +275,7 @@ int main(int argc, char **argv) {
 				} else if(arg == "-major-ring") {
 					major_ring_only = 1;
 				} else if(arg == "-no-donuts") {
-					no_donuts = 1;
+					output_no_donuts = 1;
 				} else if(arg == "-min-ring-area") {
 					if(argp == arg_list.size()) usage(cmdname);
 					min_ring_area = boost::lexical_cast<int64_t>(arg_list[argp++]);
@@ -259,6 +292,22 @@ int main(int argc, char **argv) {
 				} else if(arg == "-llproj-toler") {
 					if(argp == arg_list.size()) usage(cmdname);
 					llproj_toler = boost::lexical_cast<double>(arg_list[argp++]);
+				} else if(arg == "-containing" || arg == "-not-containing") {
+					if(argp+3 > arg_list.size()) usage(cmdname);
+					ContainingOption opt;
+					opt.wanted_point = (arg == "-containing");
+
+					std::string cs = arg_list[argp++];
+					if     (cs == "xy") opt.cs = CS_XY;
+					else if(cs == "en") opt.cs = CS_EN;
+					else if(cs == "ll") opt.cs = CS_LL;
+					else if(cs == "percent") opt.cs = CS_PERCENT;
+					else fatal_error("unrecognized coordinate system for -containing option (%s)", cs.c_str());
+
+					opt.x = boost::lexical_cast<double>(arg_list[argp++]);
+					opt.y = boost::lexical_cast<double>(arg_list[argp++]);
+
+					containing_options.push_back(opt);
 				} else if(arg == "-h" || arg == "--help") {
 					usage(cmdname);
 				} else {
@@ -277,12 +326,10 @@ int main(int argc, char **argv) {
 
 	bool do_geom_output = geom_outputs.size();
 
-	if(major_ring_only && min_ring_area) fatal_error(
-		"-major-ring and -min-ring-area options cannot both be used at the same time");
-	if(major_ring_only && no_donuts) fatal_error(
+	if(major_ring_only && output_no_donuts) fatal_error(
 		"-major-ring and -no-donuts options cannot both be used at the same time");
 
-	if(do_pinch_excursions && !(no_donuts || major_ring_only)) {
+	if(do_pinch_excursions && !(output_no_donuts || major_ring_only)) {
 		// some extra logic would be needed in pinch_excursions2 in order to
 		// support holes
 		fatal_error("the -pinch-excursions option requires the -no-donuts or -major-ring options");
@@ -327,8 +374,8 @@ int main(int argc, char **argv) {
 
 	DebugPlot *dbuf = NULL;
 	if(debug_report.size()) {
-		dbuf = new DebugPlot(georef.w, georef.h);
-		dbuf->mode = do_pinch_excursions ? PLOT_PINCH : PLOT_CONTOURS;
+		dbuf = new DebugPlot(georef.w, georef.h,
+			do_pinch_excursions ? PLOT_PINCH : PLOT_CONTOURS);
 	}
 
 	std::vector<uint8_t> raster;
@@ -384,8 +431,6 @@ int main(int argc, char **argv) {
 				(split_polys ? wkbPolygon : wkbMultiPolygon), NULL);
 			if(!go.ogr_layer) fatal_error("cannot create OGR layer");
 
-			go.class_fld_idx = -1;
-			for(int i=0; i<4; i++) go.color_fld_idx[i] = -1;
 			if(classify) {
 				OGRFieldDefnH fld = OGR_Fld_Create("value", OFTInteger);
 				OGR_Fld_SetWidth(fld, 4);
@@ -433,10 +478,62 @@ int main(int argc, char **argv) {
 			mask.erode();
 		}
 
-		Mpoly feature_poly = calc_ring_from_mask(mask, georef.w, georef.h,
-			major_ring_only, no_donuts, min_ring_area, bevel_size);
+		if(!containing_options.empty()) {
+			// We need to trace donuts even if not outputting them, in order to
+			// see if the polygons satisfy the containment options.  Ideally
+			// the user should be able to specify which happens first, hole
+			// removal or containment options.  Maybe there needs to be a
+			// rudimentary scripting language?  Or maybe just process the
+			// options in the order they are specified on the command line.
+
+			// Note: in this case, donuts must be removed later on!
+			trace_no_donuts = 0;
+		} else {
+			trace_no_donuts = output_no_donuts;
+			// If taking only the major ring, no holes are needed.
+			trace_no_donuts |= major_ring_only;
+		}
+		// If we are only taking the largest ring, and don't need to compute
+		// containments, then skip donuts for speed.
+		if(major_ring_only && containing_options.empty()) {
+			trace_no_donuts = 1;
+		}
+
+		Mpoly feature_poly = trace_mask(mask, georef.w, georef.h, min_ring_area, trace_no_donuts);
 		mask = BitGrid(0, 0); // free some memory
 
+		if(VERBOSE) {
+			size_t num_inner = 0, num_outer = 0, total_pts = 0;
+			for(size_t r_idx=0; r_idx<feature_poly.rings.size(); r_idx++) {
+				if(feature_poly.rings[r_idx].is_hole) num_inner++;
+				else num_outer++;
+				total_pts += feature_poly.rings[r_idx].pts.size();
+			}
+			printf("tracer produced %zd rings (%zd outer, %zd holes) with a total of %zd points\n",
+				feature_poly.rings.size(), num_outer, num_inner, total_pts);
+		}
+
+		if(!feature_poly.rings.empty() && !containing_options.empty()) {
+			feature_poly = containment_filters(feature_poly, containing_options, georef, dbuf);
+		}
+
+		if(major_ring_only && feature_poly.rings.size() > 1) {
+			printf("Taking largest ring.\n");
+			feature_poly = take_largest_ring(feature_poly);
+		}
+
+		if(output_no_donuts && !trace_no_donuts) {
+			// Hole removal was deferred until now.
+			printf("Removing donut holes.\n");
+			feature_poly = remove_holes(feature_poly);
+		}
+
+		if(!feature_poly.rings.empty() && bevel_size > 0) {
+			// the topology cannot be resolved by us or by geos/jump/postgis if
+			// there are self-intersections
+			bevel_self_intersections(feature_poly, bevel_size);
+		}
+
 		if(feature_poly.rings.size() && do_pinch_excursions) {
 			printf("Pinching excursions...\n");
 			feature_poly = pinch_excursions2(feature_poly, dbuf);
@@ -452,7 +549,9 @@ int main(int argc, char **argv) {
 			feature_poly = reduced_poly;
 		}
 
-		if(feature_poly.rings.size()) {
+		if(feature_poly.rings.empty()) {
+			printf("WARNING! No rings found!\n");
+		} else {
 			size_t num_outer=0, num_inner=0, total_pts=0;
 			for(size_t r_idx=0; r_idx<feature_poly.rings.size(); r_idx++) {
 				if(feature_poly.rings[r_idx].is_hole) num_inner++;
@@ -551,107 +650,174 @@ int main(int argc, char **argv) {
 		else printf("Wrote empty shapefile.\n");
 	}
 
+	GDALClose(ds);
+
 	CPLPopErrorHandler();
 
 	return 0;
 }
 
-Mpoly calc_ring_from_mask(BitGrid &mask, size_t w, size_t h,
-bool major_ring_only, bool no_donuts, 
-int64_t min_ring_area, double bevel_size) {
-	if(major_ring_only) no_donuts = 1;
-
-	Mpoly mp = trace_mask(mask, w, h, min_ring_area, no_donuts);
-
-	if(VERBOSE) {
-		size_t num_inner = 0, num_outer = 0, total_pts = 0;
-		for(size_t r_idx=0; r_idx<mp.rings.size(); r_idx++) {
-			if(mp.rings[r_idx].is_hole) num_inner++;
-			else num_outer++;
-			total_pts += mp.rings[r_idx].pts.size();
+Mpoly take_largest_ring(const Mpoly &mp_in) {
+	double biggest_area = 0;
+	size_t best_idx = 0;
+	for(size_t i=0; i<mp_in.rings.size(); i++) {
+		double area = mp_in.rings[i].area();
+		if(area > biggest_area) {
+			biggest_area = area;
+			best_idx = i;
 		}
-		printf("tracer produced %zd rings (%zd outer, %zd holes) with a total of %zd points\n",
-			mp.rings.size(), num_outer, num_inner, total_pts);
 	}
+	if(VERBOSE) printf("major ring was %zd with %zd pts, %.1f area\n",
+		best_idx, mp_in.rings[best_idx].pts.size(), biggest_area);
+	if(mp_in.rings[best_idx].parent_id >= 0) fatal_error("largest ring should not have a parent");
 
-	// this is now done directly by tracer
-	/*
-	if(min_ring_area > 0) {
-		if(VERBOSE) printf("removing small rings...\n");
-
-		ring_t *filtered_rings = MYALLOC(ring_t, mp.rings.size());
-		int *parent_map = MYALLOC(int, mp.rings.size());
-		for(int i=0; i<mp.rings.size(); i++) {
-			parent_map[i] = -1;
+	Mpoly new_mp;
+	new_mp.rings.push_back(mp_in.rings[best_idx]);
+	return new_mp;
+}
+
+Mpoly remove_holes(const Mpoly &mp_in) {
+	Mpoly new_mp;
+
+	for(size_t i=0; i<mp_in.rings.size(); i++) {
+		const Ring &ring = mp_in.rings[i];
+		// Take only top-level rings.  Since we are filling holes, it doesn't
+		// make sense to keep an island within a hole.
+		if(ring.parent_id < 0) {
+			new_mp.rings.push_back(ring);
+		}
+	}
+
+	return new_mp;
+}
+
+Mpoly containment_filters(
+	const Mpoly &mp_in,
+	const std::vector<ContainingOption> &containing_options,
+	const GeoRef &georef,
+	DebugPlot *dbuf
+) {
+	std::vector<Vertex> wanted_pts;
+	std::vector<Vertex> unwanted_pts;
+	BOOST_FOREACH(const ContainingOption &opt, containing_options) {
+		Vertex v;
+		switch(opt.cs) {
+			case CS_XY:
+				v.x = opt.x;
+				v.y = opt.y;
+				break;
+			case CS_PERCENT:
+				v.x = opt.x / 100.0 * georef.w;
+				v.y = opt.y / 100.0 * georef.h;
+				break;
+			case CS_EN:
+				georef.en2xy(opt.x, opt.y, &v.x, &v.y);
+				break;
+			case CS_LL:
+				georef.ll2xy(opt.x, opt.y, &v.x, &v.y);
+				break;
+			default:
+				fatal_error("coord system not implemented");
+		};
+		if(opt.wanted_point) {
+			wanted_pts.push_back(v);
+			dbuf->plotPointBig(v.x, v.y, 0, 255, 0);
+		} else {
+			unwanted_pts.push_back(v);
+			dbuf->plotPointBig(v.x, v.y, 255, 0, 0);
+		}
+	}
+
+	if(wanted_pts.empty() && unwanted_pts.empty()) {
+		fatal_error("no wanted/unwanted pts given");
+	}
+
+	printf("Looking for polygons");
+	if(!wanted_pts.empty()) {
+		printf(" containing:");
+		BOOST_FOREACH(const Vertex &v, wanted_pts) {
+			printf(" (%.1f,%.1f)", v.x, v.y);
+		}
+	}
+	if(!unwanted_pts.empty()) {
+		printf(" not containing:");
+		BOOST_FOREACH(const Vertex &v, unwanted_pts) {
+			printf(" (%.1f,%.1f)", v.x, v.y);
 		}
-		int num_filtered_rings = 0;
-		for(int i=0; i<mp.rings.size(); i++) {
-			double area = ring_area(mp.rings+i);
-			if(VERBOSE) if(area > 10) printf("ring %zd has area %.15f\n", i, area);
-			if(area >= min_ring_area) {
-				parent_map[i] = num_filtered_rings;
-				filtered_rings[num_filtered_rings++] = mp.rings[i];
-			} else {
-				free_ring(mp.rings + i);
+	}
+	printf("\n");
+
+	Mpoly new_mp;
+	std::map<int, int> relabeling;
+
+	int num_outer=0, num_holes=0;
+
+	for(size_t outer_idx=0; outer_idx<mp_in.rings.size(); outer_idx++) {
+		const Ring &outer = mp_in.rings[outer_idx];
+		if(outer.is_hole) continue;
+
+		bool contains_wanted_pt = false;
+		BOOST_FOREACH(const Vertex &v, wanted_pts) {
+			if(mp_in.component_contains(v, outer_idx)) {
+				if(VERBOSE) printf("ring %zd contains wanted point %g,%g\n",
+					outer_idx, v.x, v.y);
+				contains_wanted_pt = true;
+				break;
 			}
 		}
-		for(int i=0; i<num_filtered_rings; i++) {
-			int old_parent = filtered_rings[i].parent_id;
-			if(old_parent >= 0) {
-				int new_parent = parent_map[old_parent];
-				if(new_parent < 0) fatal_error("took a ring but not its parent");
-				filtered_rings[i].parent_id = new_parent;
-			}
+		if(!wanted_pts.empty() && !contains_wanted_pt) {
+			// not interested in this ring
+			continue;
 		}
-		printf("filtered by area %zd => %zd rings\n",
-			mp.rings.size(), num_filtered_rings);
 
-		free(mp.rings);
-		mp.rings = filtered_rings;
-		mp.rings.size() = num_filtered_rings;
-	}
-	*/
-
-	if(major_ring_only && mp.rings.size() > 1) {
-		double biggest_area = 0;
-		size_t best_idx = 0;
-		for(size_t i=0; i<mp.rings.size(); i++) {
-			double area = mp.rings[i].area();
-			if(area > biggest_area) {
-				biggest_area = area;
-				best_idx = i;
+		bool contains_unwanted_pt = false;
+		BOOST_FOREACH(const Vertex &v, unwanted_pts) {
+			if(mp_in.component_contains(v, outer_idx)) {
+				if(VERBOSE) printf("ring %zd contains unwanted point %g,%g\n",
+					outer_idx, v.x, v.y);
+				contains_unwanted_pt = true;
+				break;
 			}
+		}
+		if(contains_unwanted_pt) {
+			// not interested in this ring
+			continue;
 		}
-		if(VERBOSE) printf("major ring was %zd with %zd pts, %.1f area\n",
-			best_idx, mp.rings[best_idx].pts.size(), biggest_area);
-		if(mp.rings[best_idx].parent_id >= 0) fatal_error("largest ring should not have a parent");
+
+		relabeling[outer_idx] = int(new_mp.rings.size());
+		new_mp.rings.push_back(outer);
+		num_outer++;
 
-		Mpoly new_mp;
-		new_mp.rings.push_back(mp.rings[best_idx]);
-		mp = new_mp;
+		for(size_t j=0; j<mp_in.rings.size(); j++) {
+			Ring inner = mp_in.rings[j];
+			// take children of outer ring
+			if(inner.parent_id != int(outer_idx)) continue;
+
+			relabeling[j] = int(new_mp.rings.size());
+			new_mp.rings.push_back(inner);
+			num_holes++;
+		}
 	}
 
-	// this is now done directly by tracer
-	/*
-	if(no_donuts) {
-		// we don't have to worry about remapping parent_id in this
-		// case because we only take rings with no parent
-		int out_idx = 0;
-		for(int i=0; i<mp.rings.size(); i++) {
-			if(mp.rings[i].parent_id < 0) {
-				mp.rings[out_idx++] = mp.rings[i];
+	for(size_t i=0; i<new_mp.rings.size(); i++) {
+		Ring &ring = new_mp.rings[i];
+		// Compute new parent.  Iterate outwards through the ring hierarchy
+		// until a ring is found that has been kept.
+		while(ring.parent_id >= 0) {
+			if(relabeling.count(ring.parent_id)) {
+				ring.parent_id = relabeling[ring.parent_id];
+				break;
+			} else {
+				ring.parent_id = mp_in.rings[ring.parent_id].parent_id;
 			}
 		}
-		mp.num_rings = out_idx;
-		if(VERBOSE) printf("number of non-donut rings is %zd", mp.rings.size());
 	}
-	*/
 
-	if(mp.rings.size() && bevel_size > 0) {
-		// the topology cannot be resolved by us or by geos/jump/postgis if
-		// there are self-intersections
-		bevel_self_intersections(mp, bevel_size);
+	if(new_mp.rings.empty()) {
+		printf("   None found!\n");
+	} else {
+		printf("   Found %d connected components (with %d holes).\n", num_outer, num_holes);
 	}
 
-	return mp;
+	return new_mp;
 }
