@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2009, Regents of the University of Alaska
+Copyright (c) 2013, Regents of the University of Alaska
 
 All rights reserved.
 
@@ -70,7 +70,7 @@ void Bbox::expand(const Bbox &bb) {
 	}
 }
 
-Bbox box_union(const Bbox bb1, const Bbox bb2) {
+Bbox box_union(const Bbox &bb1, const Bbox &bb2) {
 	Bbox u = bb1;
 	u.expand(bb2);
 	return u;
@@ -103,7 +103,7 @@ Ring ogr_to_ring(OGRGeometryH ogr) {
 	return ring;
 }
 
-OGRGeometryH mpoly_to_ogr(const Mpoly mpoly_in) {
+OGRGeometryH mpoly_to_ogr(const Mpoly &mpoly_in) {
 	size_t num_rings_in = mpoly_in.rings.size();
 
 	std::vector<std::vector<int> > holes;
@@ -416,6 +416,20 @@ bool Mpoly::contains(Vertex p) const {
 	return num_crossings & 1;
 }
 
+bool Mpoly::component_contains(Vertex p, int outer_ring_id) const {
+	const Ring &outer = rings[outer_ring_id];
+	if(outer.is_hole) fatal_error("ring was a hole in Mpoly::component_contains");
+	if(!outer.contains(p)) return false;
+
+	for(size_t j=0; j<rings.size(); j++) {
+		const Ring &inner = rings[j];
+		// take children of outer ring
+		if(inner.parent_id != int(outer_ring_id)) continue;
+		if(inner.contains(p)) return false;
+	}
+	return true;
+}
+
 void Mpoly::deleteRing(size_t idx) {
 	rings.erase(rings.begin() + idx);
 }
@@ -814,6 +828,7 @@ Mpoly mpoly_from_wktfile(const std::string &fn) {
 	FILE *fh = fopen(fn.c_str(), "r");
 	if(!fh) fatal_error("cannot read file [%s]", fn.c_str());
 	std::string wkt = read_whole_file(fh);
+	fclose(fh);
 	for(size_t i=0; i<wkt.size(); i++) {
 		if(wkt[i] == '\r') wkt[i] = ' ';
 		if(wkt[i] == '\n') wkt[i] = ' ';
